unit CustomerList;

interface

uses
  System.Collections,
  System.ComponentModel,
  Customer,
  BugLists;

type

  // Forward declaration
  TCustomerListEnumerator = class;

{$REGION 'Main collection class'}
  TCustomerList = class( System.Collections.CollectionBase,
                         System.ComponentModel.IBindingList,
                         System.Collections.IEnumerable )
  strict private

    _listChanged: ListChangedEventHandler;

    procedure OnListChanged( p_sender: TObject; p_eventArgs: ListChangedEventArgs );

  private

    {$REGION 'Filtering'}

    _lastNameBeginsWith: string;

    {$ENDREGION}

    {$REGION 'Generate test data'}

    _titleArray: array [ 0..2 ] of string;
    _lastNameArray: array [ 0..4 ] of string;
    _random: System.Random;

    {$ENDREGION}

    {$REGION 'Generate test data'}

    procedure LoadDataArrays;

    {$ENDREGION}

    {$REGION 'IList Property getters & setters'}
    function get_FixedSize: boolean;
    function get_IsReadOnly: boolean;
    function get_Item(p_itemNo: integer): TObject;
    procedure set_Item(p_itemNo: integer; p_customer: TObject);
    {$ENDREGION}

    {$REGION 'IList members'}
    function Add( p_newCustomer: TObject ): integer; overload;
    // procedure Clear;     // not needed - implemented by CollectionBase
    function Contains( p_customer: TObject ): boolean; overload;
    function IndexOf( p_customer: TObject ): integer; overload;
    procedure Insert( p_position: integer; p_customer: TObject ); overload;
    procedure Remove( p_customer: TObject ); overload;
    procedure RemoveAt( p_position: integer ); reintroduce;

    property IsFixedSize: boolean read get_FixedSize;
    property IsReadOnly: boolean read get_IsReadOnly;
    property Item[ p_itemNo: integer ]: TObject read get_Item write set_Item; default;
    {$ENDREGION}

  strict protected

    {$REGION 'CollectionBase event overrides'}
    procedure OnValidate( value: TObject ); override;
    procedure OnInsertComplete( index: integer; value: TObject ); override;
    {$ENDREGION}

  public

    {$REGION 'IBindingList Property getters & setters'}
    function get_AllowEdit: boolean;                    // return True
    function get_AllowNew: boolean;                     // return True
    function get_AllowRemove: boolean;                  // return True
    function get_IsSorted: boolean;                     // return False
    function get_SortDirection: ListSortDirection;      // raise NotSupportedException
    function get_SortProperty: PropertyDescriptor;      // raise NotSupportedException
    function get_SupportsChangeNotification: boolean;   // return True
    function get_SupportsSearching: boolean;            // return False
    function get_SupportsSorting: boolean;              // return False
    {$ENDREGION}

  public

    {$REGION 'Strongly typed IList methods'}
    function Add( p_newCustomer: ICustomer ): integer; overload;
    function Contains( p_customer: ICustomer ): boolean; overload;
    function IndexOf( p_customer: ICustomer ): integer; overload;
    procedure Insert( p_position: integer; p_customer: ICustomer ); overload;
    procedure Remove( p_customer: ICustomer ); overload;
    {$ENDREGION}

    {$REGION 'ICollection members'}
    procedure CopyTo( p_array: System.Array; p_index: integer );
    function get_SyncRoot: TObject;
    function get_IsSynchronized: boolean;
    {$ENDREGION}

    {$REGION 'Sorting'}

    procedure Sort; overload;
    procedure Sort( p_comparer: IComparer ); overload;
    procedure Reverse;
    procedure SortByLastName;

    {$ENDREGION}

    {$REGION 'Filtering'}

    property LastNameBeginsWith: string read _lastNameBeginsWith write _lastNameBeginsWith;

    {$ENDREGION}

  public

    {$REGION 'Constructor'}
    constructor Create;
    {$ENDREGION}

    {$REGION 'IEnumerable members'}

    function GetEnumerator: IEnumerator; reintroduce;

    {$ENDREGION}

    {$REGION 'Generate test data'}

    procedure GenerateTestData( p_noOfCustomers: integer );
    procedure InitialiseCustomer( p_customer: ICustomer; p_customerNo: integer );

    {$ENDREGION}

    {$REGION 'IBindingList methods'}
    procedure AddIndex( p_property: PropertyDescriptor );       // Do nothing
    function AddNew: TObject;
    procedure ApplySort( p_property: PropertyDescriptor;        // raise NotSupportedException
                                                p_direction: ListSortDirection );
    function Find( p_property: PropertyDescriptor; p_key: TObject ): integer; // raise NotSupportedException
    procedure RemoveIndex( p_property: PropertyDescriptor );                  // Do nothing
    procedure RemoveSort;                                                     // raise NotSupportedException
    property ListChanged: ListChangedEventHandler add _listChanged remove _listChanged;
    {$ENDREGION}

  end;
{$ENDREGION}

{$REGION 'Enumerator'}
  TCustomerListEnumerator = class( TObject, System.Collections.IEnumerator )
  strict private

  {$REGION 'Enumeration variables'}
    _position: integer;
    _customers: TCustomerList;
  {$ENDREGION}

  {$REGION 'Filtering variables'}
    _lastNameBeginsWith: string;
  {$ENDREGION}

  {$REGION 'Helper function'}
    function CurrentCustomer: TCustomer;
  {$ENDREGION}

  strict protected

  {$REGION 'IEnumerator members'}
    function MoveNext: boolean;
    procedure Reset;

    function get_Current: TObject;

    property Current: TObject read get_Current;
  {$ENDREGION}

  public

  {$REGION 'Constructor'}
    constructor Create( p_customers: TCustomerList; p_lastNameBeginsWith: string );
  {$ENDREGION}

  end;
{$ENDREGION}

implementation

{$REGION 'Main collection class'}

{$REGION 'IList overrides' }
function TCustomerList.Add(p_newCustomer: TObject): integer;
begin
  if ( p_newCustomer is ICustomer ) then
  begin
    Result := InnerList.Add( p_newCustomer as TObject );
    OnListChanged( Self, ListChangedEventArgs.Create( ListChangedType.ItemAdded, Result ) );
  end
  else
  begin
    raise ArgumentException.Create( 'You may only add an ICustomer to a TCustomerList' );
  end;
end;

function TCustomerList.Contains(p_customer: TObject): boolean;
begin
  if ( p_customer is ICustomer ) then
  begin
    Result := InnerList.Contains( p_customer as ICustomer );
  end
  else
  begin
    raise ArgumentException.Create( 'You may only check if a TCustomerList contains an ICustomer' );
  end;
end;

function TCustomerList.get_FixedSize: boolean;
begin
  Result := InnerList.IsFixedSize;
end;

function TCustomerList.get_IsReadOnly: boolean;
begin
  Result := InnerList.IsReadOnly;
end;

function TCustomerList.get_Item(p_itemNo: integer): TObject;
begin
  Result := InnerList.Item[ p_itemNo ] as TCustomer;
end;

function TCustomerList.IndexOf(p_customer: TObject): integer;
begin
  if ( p_customer is ICustomer ) then
  begin
    Result := InnerList.IndexOf( p_customer );
  end
  else
  begin
    raise ArgumentException.Create( 'You may only get an IndexOf an ICustomer from a TCustomerList' );
  end;
end;

procedure TCustomerList.Insert(p_position: integer; p_customer: TObject);
begin
  if ( p_customer is ICustomer ) then
  begin
    InnerList.Insert( p_position, p_customer );
    OnListChanged( Self, ListChangedEventArgs.Create( ListChangedType.ItemAdded, p_position ) );
  end
  else
  begin
    raise ArgumentException.Create( 'You may only insert an ICustomer into a TCustomerList' );
  end;
end;

procedure TCustomerList.Remove(p_customer: TObject);
var
  l_position: integer;
begin
  if ( p_customer is ICustomer ) then
  begin
    l_position := InnerList.IndexOf( p_customer );
    if ( l_position > -1 ) then
    begin
      Self.RemoveAt( l_position );    // Will also fire the event
    end;
  end
  else
  begin
    raise ArgumentException.Create( 'You may only remove an ICustomer from a TCustomerList' );
  end;
end;

procedure TCustomerList.RemoveAt(p_position: integer);
begin
  InnerList.RemoveAt( p_position );
  OnListChanged( Self, ListChangedEventArgs.Create( ListChangedType.ItemDeleted, p_position ) );
end;

procedure TCustomerList.set_Item(p_itemNo: integer;
  p_customer: TObject);
begin
  if ( p_customer is ICustomer ) then
  begin
    InnerList.Item[ p_itemNo ] := p_customer;
    OnListChanged( Self, ListChangedEventArgs.Create( ListChangedType.ItemChanged, p_itemNo ) );
  end
  else
  begin
    raise ArgumentException.Create( 'You may only put an ICustomer into a TCustomerList' );
  end;
end;
{$ENDREGION}

{$REGION 'IEnumerable members'}
function TCustomerList.GetEnumerator: IEnumerator;
begin
  Result := TCustomerListEnumerator.Create( Self, _lastNameBeginsWith );
end;
{$ENDREGION}

{$REGION 'Sorting'}
procedure TCustomerList.Sort;
begin
  Self.InnerList.Sort;
  OnListChanged( Self, ListChangedEventArgs.Create( ListChangedType.Reset, 0 ) );
end;

procedure TCustomerList.Reverse;
begin
  Self.InnerList.Reverse;
  OnListChanged( Self, ListChangedEventArgs.Create( ListChangedType.Reset, 0 ) );
end;

procedure TCustomerList.Sort(p_comparer: IComparer);
begin
  Self.InnerList.Sort( p_comparer );
  OnListChanged( Self, ListChangedEventArgs.Create( ListChangedType.Reset, 0 ) );
end;
{$ENDREGION}

{$REGION 'Utility procedures'}
procedure TCustomerList.SortByLastName;
var
  l_comparer: TCompareCustomerByLastName;
begin
  l_comparer := TCompareCustomerByLastName.Create;
  try
    Self.Sort( l_comparer );
  finally
    l_comparer.Free;
  end;
end;

procedure TCustomerList.GenerateTestData( p_noOfCustomers: integer );
var
  l_customerNo: integer;
  l_newCustomer: TCustomer;
begin
  for l_customerNo := 0 to Pred( p_noOfCustomers ) do
  begin
    l_newCustomer := TCustomer.Create( Self );
    l_newCustomer.BeginEdit;
    Self.InitialiseCustomer( l_newCustomer, l_customerNo );
    Self.Add( l_newCustomer );
    l_newCustomer.EndEdit;
  end;
end;

procedure TCustomerList.InitialiseCustomer(p_customer: ICustomer;
  p_customerNo: integer);
begin
  p_customer.Title := _titleArray[ _random.Next( 3 ) ];
  p_customer.FirstName := 'Customer ' + p_customerNo.ToString;
  p_customer.LastName := _lastNameArray[ _random.Next( 5 ) ];
end;

procedure TCustomerList.LoadDataArrays;
begin
  _titleArray[ 0 ] := 'Mr';
  _titleArray[ 1 ] := 'Mrs';
  _titleArray[ 2 ] := 'Miss';
  _lastNameArray[ 0 ] := 'Smith';
  _lastNameArray[ 1 ] := 'Jones';
  _lastNameArray[ 2 ] := 'Brown';
  _lastNameArray[ 3 ] := 'Green';
  _lastNameArray[ 4 ] := 'Cooper';
end;
{$ENDREGION}

{$REGION 'Constructor'}
constructor TCustomerList.Create;
begin
  inherited Create;
  if ( not Assigned( _random ) ) then
  begin
    _random := System.Random.Create;
  end;
  LoadDataArrays;
  _lastNameBeginsWith := '';
end;
{$ENDREGION}

{$REGION 'IBindingList property getters & setters'}
function TCustomerList.get_AllowEdit: boolean;
begin
  Result := True;
end;

function TCustomerList.get_AllowNew: boolean;
begin
  Result := True;
end;

function TCustomerList.get_AllowRemove: boolean;
begin
  Result := True;
end;

function TCustomerList.get_IsSorted: boolean;
begin
  Result := False;
end;

function TCustomerList.get_SortDirection: ListSortDirection;
begin
  raise NotSupportedException.Create;
end;

function TCustomerList.get_SortProperty: PropertyDescriptor;
begin
  raise NotSupportedException.Create;
end;

function TCustomerList.get_SupportsChangeNotification: boolean;
begin
  Result := True;
end;

function TCustomerList.get_SupportsSearching: boolean;
begin
  Result := False;
end;

function TCustomerList.get_SupportsSorting: boolean;
begin
  Result := False;
end;
{$ENDREGION}

{$REGION 'IBindingList methods'}
procedure TCustomerList.AddIndex(p_property: PropertyDescriptor);
begin
  // Do nothing
end;

function TCustomerList.AddNew: TObject;
var
  l_newCustomer: TCustomer;
begin
  l_newCustomer := TCustomer.Create( Self );
  Self.Add( l_newCustomer );
  Result := l_newCustomer;
end;

procedure TCustomerList.ApplySort(p_property: PropertyDescriptor;
  p_direction: ListSortDirection);
begin
  // Not supported
  raise NotSupportedException.Create;
end;

function TCustomerList.Find(p_property: PropertyDescriptor;
  p_key: TObject): integer;
begin
  // Not supported yet
  raise NotSupportedException.Create;
end;

procedure TCustomerList.RemoveIndex(p_property: PropertyDescriptor);
begin
  // Do nothing
end;

procedure TCustomerList.RemoveSort;
begin
  // Not supported yet
  raise NotSupportedException.Create;
end;
{$ENDREGION}

{$REGION 'Helper method for IBindingList event handling'}
procedure TCustomerList.OnListChanged( p_sender: TObject; p_eventArgs: ListChangedEventArgs );
begin
  if ( Assigned( _listChanged ) ) then
  begin
    _listChanged( p_sender, p_eventArgs );
  end;
end;
{$ENDREGION}

{$REGION 'ICollection members'}
procedure TCustomerList.CopyTo(p_array: System.Array; p_index: integer);
begin
  ( ( Self as CollectionBase ) as ICollection ).CopyTo( p_array, p_index );
end;

function TCustomerList.get_SyncRoot: TObject;
begin
  Result := ( ( Self as CollectionBase ) as ICollection ).SyncRoot;
end;

function TCustomerList.get_IsSynchronized: boolean;
begin
  Result := ( ( Self as CollectionBase ) as ICollection ).IsSynchronized;
end;
{$ENDREGION}

{$REGION 'Strongly typed list management methods'}
function TCustomerList.Add(p_newCustomer: ICustomer): integer;
begin
  Result := List.Add( p_newCustomer );
end;

function TCustomerList.Contains(p_customer: ICustomer): boolean;
begin
  Result := List.Contains( p_customer );
end;

function TCustomerList.IndexOf(p_customer: ICustomer): integer;
begin
  Result := List.IndexOf( p_customer );
end;

procedure TCustomerList.Insert(p_position: integer; p_customer: ICustomer);
begin
  List.Insert( p_position, p_customer );
end;

procedure TCustomerList.Remove(p_customer: ICustomer);
begin
  List.Remove( p_customer );
end;
{$ENDREGION}
{$ENDREGION}

{$REGION 'Enumerator'}
{ TCustomerListEnumerator }

{$REGION 'Enumerator Constructor'}
constructor TCustomerListEnumerator.Create( p_customers: TCustomerList;
        p_lastNameBeginsWith: string );
begin
  inherited Create;
  _customers := p_customers;
  _position := -1;
  _lastNameBeginsWith := p_lastNameBeginsWith;
end;
{$ENDREGION}

{$REGION 'Helper function'}
function TCustomerListEnumerator.CurrentCustomer: TCustomer;
begin
  Result := Self._customers[ _position ] as TCustomer;
end;
{$ENDREGION}

{$REGION 'IEnumerator members'}
function TCustomerListEnumerator.get_Current: TObject;
begin
  Result := Self.CurrentCustomer;
end;

function TCustomerListEnumerator.MoveNext: boolean;
var
  l_oldPosition: integer;
begin
  // Test to see if there is another element of the list after the current one
  if ( Self._position < ( Self._customers.Count - 1 ) ) then
  begin
    l_oldPosition := Self._position;
    Self._position := Self._position + 1;
    // Do filtering, if required
    if ( _lastNameBeginsWith.Length > 0 ) then
    begin
      // Check the start of the Last Name against the comparison string (case insensitive)
      // If they don't match, move to the next Customer, until we reach the end of the list
      while ( ( Self._position < Self._customers.Count ) and
              ( System.string.Compare( Self.CurrentCustomer.LastName,
                                       0,
                                       _lastNameBeginsWith,
                                       0,
                                       _lastNameBeginsWith.Length,
                                       true ) <> 0 ) ) do
      begin
        Self._position := Self._position + 1;
      end;
    end;
    Result := _position < Self._customers.Count;
    if ( not Result ) then
    begin
      // If there isn't another Customer that meets the filtering criteria,
      // restore the old position
      Self._position := l_oldPosition;
    end;
  end
  else
  begin
    Result := False;
  end;
end;

procedure TCustomerListEnumerator.Reset;
begin
  Self._position := -1;
end;
{$ENDREGION}
{$ENDREGION}

{$REGION 'CollectionBase event overrides'}
procedure TCustomerList.OnInsertComplete(index: integer; value: TObject);
begin

end;

procedure TCustomerList.OnValidate(value: TObject);
begin
  if ( not ( value is ICustomer ) ) then
  begin
    raise ArgumentException.Create( 'You may only put an ICustomer into a TCustomerList collection' );
  end;
end;
{$ENDREGION}

end.
